import os, json


def open_json(path: str, default = {}) -> dict:
    """
    Attempts to open a json file and returns its contents as a dict.
    If the file does not exist, a new one will be created with 'default'
    """
    if not os.path.exists(path):
        with open(path, "x"):
            pass
    with open(path, "r") as file:
        content = file.read()
        if content.strip():
            data = json.loads(content)
        else:
            data = default
    file.close()
    return data


def save_json(data: dict, path: str) -> None:
    """Saves a json file"""
    with open(path, "w") as file:
        json.dump(data, file)
    file.close()





import cProfile
import pstats
import psutil
import os
from io import StringIO
import inspect
import functools




def profile_run(command: str, sort_by: str = 'cumtime', lines = None, precision = 6) -> None:
    """
    A custom profiling function similar to cProfile.run() with enhanced formatting options.

    Parameters:
    - command (str): The string of code to execute (must be a valid Python expression).
    - sort_by (str): The metric to sort the profiling results by. Options: 'tottime', 'cumtime', 'ncalls', etc.
    - lines (int): Number of lines of stats to display.
    - precision (int): Number of decimal places to display for timing statistics.
    """
    # Get the caller's frame
    caller_frame = inspect.stack()[1]

    # Extract globals from the caller's frame (where the function is defined)
    caller_globals = caller_frame[0].f_globals

    #Track process
    process = psutil.Process(os.getpid())

    #Get initial memory usage (in bytes)
    mem_initial = process.memory_info().rss

    #Run Command
    profiler = cProfile.Profile()
    profiler.enable()
    exec(command, caller_globals, locals())
    profiler.disable()

    mem_final = process.memory_info().rss
    memory_usage = (mem_final - mem_initial) / (1024**2) #MB

    # Capture profiling stats
    result_stream = StringIO()
    stats = pstats.Stats(profiler, stream=result_stream)
    stats.strip_dirs().sort_stats(sort_by)

    #Print header
    header = '|' + f"  Profile for Command:  \"{command}\"  " + '|'
    print('~' * len(header))
    print(header)
    print('~' * len(header), '\n')

    #Print all lines if lines is None
    if lines:
        stats.print_stats(lines)
    else:
        stats.print_stats()

    # Customize precision in the output
    formatted_output = result_stream.getvalue()
    # Adjust decimal places by formatting lines with timing data
    lines = formatted_output.split('\n')

    for i, line in enumerate(lines):
        # Skip header lines
        if line.strip() and line[0].isdigit():
            parts = line.split()

            # Ensure timing columns are reformatted with desired precision
            if len(parts) > 3:  # ncalls, tottime, percall, cumtime, percall, filename
                for j in range(1, 5):  # Only modify timing-related columns
                    parts[j] = f'{float(parts[j]):.{precision}f}'
                lines[i] = ' '.join(parts)

    #Total Memory
    print(f"   Total Memory Usage: {memory_usage} MB\n")
    #Readout
    print('\n'.join(lines))



def preserve_signature(decorator):
    '''Preserves metadata for decorated functions.'''
    @functools.wraps(decorator)
    def wrapper(func):
        decfunc = decorator(func)

        signature = inspect.signature(func)
        decfunc.__signature__ = signature

        functools.update_wrapper(decfunc, func)

        return decfunc
    return wrapper



from pickle import dumps, PicklingError
from types import MethodType


def hashable_repr(obj):
    """Create a recursively hashable representation of any object."""
    if isinstance(obj, (str, int, float, bool, type(None), bytes)):
        return obj
    # Use specific type checks and prepend a type identifier
    elif isinstance(obj, list):
        return ('__list__',) + tuple(hashable_repr(i) for i in obj)
    elif isinstance(obj, tuple):
        return ('__tuple__',) + tuple(hashable_repr(i) for i in obj)
    elif isinstance(obj, dict):
        # Sort items to make dict representation deterministic
        return ('__dict__',) + tuple(sorted(((hashable_repr(k), hashable_repr(v)) for k, v in obj.items()), key=lambda x: str(x[0])))
    elif isinstance(obj, set):
        # Sort elements to make set representation deterministic
        return ('__set__',) + tuple(sorted((hashable_repr(i) for i in obj), key=str))
    elif isinstance(obj, MethodType) or hasattr(obj, '__dict__'):
        return (str(type(obj)), id(obj))
    else:
        # Fallback for other types.
        try:
            # Try to pickle the object. `dumps` returns hashable bytes.
            # We wrap it in a tuple to avoid collision with other types.
            return ('__pickle__', dumps(obj))
        except (PicklingError, TypeError):
            # If it's not pickleable (e.g., a lambda, local function, etc.),
            # fall back to a representation using its id.
            return ('__id__', id(obj))


from typing import Any, TypeVar
from .typemacros import tupler, maybe_arg


#Custom token to break a search in a callback.
BREAK_SEARCH = object()

Iterables = TypeVar('Iterables', list, tuple, dict, set)

def _replace_value_mutable(a: Iterables, old_vals: tuple | Any, new_val,callback = None) -> Iterables:
    callback = callback if callback is not None else lambda old, new: new
    old_vals = tupler(old_vals)

    stack = [(a, [])]  # (current structure, parents)

    # Process the stack iteratively
    while stack:
        #Remove from stack
        current, parents = stack.pop()

        #Replace the current element
        if current in old_vals:
            new_value = maybe_arg(callback)(current, new_val, parents = parents)

            #Break on token
            if new_value is BREAK_SEARCH:
                return a
            
            if parents:
                #Get first parent from list
                parent = parents[0]

                #Update the parent structure with the new value
                if isinstance(parent, list):
                    parent[parent.index(current)] = new_value
                elif isinstance(parent, dict):
                    for k, v in parent.items():
                        if v == current:
                            parent[k] = new_value
                elif isinstance(parent, set):
                    parent.remove(current)
                    parent.add(new_value)
            continue

        #Append Stack for each item in the iterable and join parents list
        if isinstance(current, list):
            stack.extend((item, [current] + parents) for item in current)
        elif isinstance(current, set):
            stack.extend((item, [current] + parents) for item in current)
        elif isinstance(current, dict):
            stack.extend((v, [current] + parents) for v in current.values())

    return a


#The stack based approach is 50% faster and handles deeper recursion, but does not work with tuples.
class _replace_value_recursive():
    def __init__(self):
        self.broken = False

    def __call__(self, a: Iterables, old_vals: tuple | Any, new_val, callback=None, parents=None) -> Iterables:
        #More memory efficient to break early
        if self.broken:
            return a
        
        parents = [] if parents is None else parents # Fix mutable default argument
        callback = callback if callback is not None else lambda old, new: new
        old_vals = tupler(old_vals)

        if a in old_vals:
            new_value = maybe_arg(callback)(a, new_val, parents = parents)
            if new_value is BREAK_SEARCH:
                self.broken = True
                return a
            else:
                return new_value

        elif isinstance(a, list):
            return type(a)(self(x, old_vals, new_val, callback, [a] + parents) for x in a)
        elif isinstance(a, tuple):
            return type(a)(self(x, old_vals, new_val, callback, [a] + parents) for x in a)
        elif isinstance(a, dict):
            return type(a)({k: self(v, old_vals, new_val, callback, [a] + parents) for k, v in a.items()})
        elif isinstance(a, set):
            return type(a)(self(x, old_vals, new_val, callback, [a] + parents) for x in a)
        else:
            return a


from types import GeneratorType
from typing import Any, TypeVar
import inspect
from .hashmacros import hashable_repr


Iterables = TypeVar('Iterables', list, tuple, dict, set)

#Unions
#----------------------------------------------------------------------------------------------------------------------
def list_union(A: list | tuple | list[list] | tuple[list], B: list | tuple | None = None) -> list:
    """
    If 'b' is None, the function will perform an iterative union of all lists and tuples in 'a'.
    In this case, 'a' must be a list or tuple of exclusively lists or  tuples.
    Otherwise, perform the union of 'a' and 'b'
    """
    if B is None:
        l_union = list()
        for n in A:
            if isinstance(n, list | tuple):
                l_union = list_union(l_union, n)
            else:
                raise TypeError("'a' must be a list or tuple of lists or tuples if 'b' is empty")
        return l_union
    else:
        seen = set(map(hashable_repr, A))
        result = list_union(list(map(lambda a: [a], A))) # Makes sure we dedup A
        for n in B:
            hash_n = hashable_repr(n)
            if hash_n not in seen:
                seen.add(hash_n)
                result.append(n)
        return result


def tuple_union(A: list | tuple | list[tuple] | tuple[tuple], B: list | tuple | None = None) -> tuple:
    """
    If 'b' is None, the function will perform an iterative union of all lists and tuples in 'a'.
    In this case, 'a' must be a list or tuple of exclusively lists or  tuples.
    Otherwise, perform the union of 'a' and 'b'
    """
    if B is None:
        t_union = ()
        for n in A:
            if isinstance(n, list | tuple):
                t_union = tuple_union(t_union, n)
            else:
                raise TypeError("'a' must be a list or tuple of lists or tuples if 'b' is empty")
        return t_union
    else:
        seen = set(map(hashable_repr, A))
        result = list_union(list(map(lambda a: [a], A))) # Makes sure we dedup A
        for n in B:
            hash_n = hashable_repr(n)
            if hash_n not in seen:
                seen.add(hash_n)
                result.append(n)
        return tuple(result)


def dict_union(A: dict | list[dict] | tuple[dict], B: dict | None = None) -> dict:
    """
    If 'b' is None, the function will perform an iterative union of all dicts in 'a'.
    In this case, 'a' must be a list or tuple of exclusively dicts.
    Otherwise, perform the union of 'a' and 'b'
    (Leftmost keys have precedence)
    """
    if B is None:
        d_union = dict()
        for n in A:
            if isinstance(n, dict):
                d_union = dict_union(d_union, n)
            else:
                raise TypeError("'a' must be a list or tuple of dicts if 'b' is empty")
        return d_union
    else:
        if isinstance(A, (list, tuple)):
            raise TypeError("'a' must be a list or tuple of dicts if 'b' is empty")
        return dict(A | B)


def set_union(A: set | list[set] | tuple[set], B: set | None = None) -> set:
    """
    If 'b' is None, the function will perform an iterative union of all sets in 'a'.
    In this case, 'a' must be a list or tuple of exclusively sets.
    Otherwise, perform the union of 'a' and 'b'
    """
    if B is None:
        s_union = set()
        for n in A:
            if isinstance(n, set):
                s_union = set_union(s_union, n)
            else:
                raise TypeError("'a' must be a list or tuple of sets if 'b' is empty")
        return s_union
    else:
        if isinstance(A, (list, tuple)):
            raise TypeError("'a' must be a list or tuple of sets if 'b' is empty")
        return set(A | B)


def type_union(A: Iterables , B: Iterables | None = None) -> Iterables:
    """General union for all types.  A and B must be the same type unless B is None
    In this case, 'A' must be a list or tuple of exclusively sets.
    Otherwise, perform the union of 'a' and 'b'
    """
    if type(A) != type(B) and B is not None:
        raise TypeError("A and B must be the same type!")
    if isinstance(A, list) and (isinstance(B, list) or B is None):
        return list_union(A, B)
    elif isinstance(A, tuple) and (isinstance(B, tuple) or B is None):
        return tuple_union(A, B)
    elif isinstance(A, dict) and (isinstance(B, dict) or B is None):
        return dict_union(A, B)
    elif isinstance(A, set) and (isinstance(B, set) or B is None):
        return set_union(A, B)
    else:
        raise TypeError(f"type_union(A: ${type(A)}, B: ${type(B)})\nWrong type signature")


#Compliments
#-------------------------------------------------------------------------------------------------------------------

def list_compliment(A: list, B: list) -> list:
    """Returns a list containing all items in A that are not in B"""
    seen = set(map(hashable_repr, B))
    return [n for n in A if hashable_repr(n) not in seen]

def tuple_compliment(A: tuple, B: tuple) -> tuple:
    """Returns a tuple containing all items in A that are not in B"""
    seen = set(map(hashable_repr, B))
    return tuple(n for n in A if hashable_repr(n) not in seen)

def dict_compliment(A: dict, B: dict, match_vals= False) -> dict:
    """Returns a dict containing all keys in A that are not in B"""
    return dict({key:val for key, val in A.items() if (key, val) not in B.items()}) if match_vals else \
        dict({key:A[key] for key in A if key not in B})

def set_compliment(A: set, B: set) -> set:
    """Returns a set containing all members in A that are not in B"""
    return set(a for a in A if a not in B)

def type_compliment(A: Iterables, B: Iterables) -> Iterables:
    """General compliment for all types.  A and B must be the same type"""
    if type(A) != type(B):
        raise TypeError("A and B must be the same type!")
    if isinstance(A, list) and isinstance(B, list):
        return list_compliment(A, B)
    elif isinstance(A, tuple) and isinstance(B, tuple):
        return tuple_compliment(A, B)
    elif isinstance(A, dict) and isinstance(B, dict):
        return dict_compliment(A, B)
    elif isinstance(A, set) and isinstance(B, set):
        return set_compliment(A, B)
    else:
        raise TypeError(f"type_compliment(A: ${type(A)}, B: ${type(B)})\nWrong type signature")


#Intersections
#---------------------------------------------------------------------------------------------------------------------

def list_intersect(A: list, B: list) -> list:
    """Returns a list returning all items in A that are also in B"""
    seen = set(map(hashable_repr, B))
    return [n for n in A if hashable_repr(n) in seen]

def tuple_intersect(A: tuple, B: tuple) -> tuple:
    """Returns a list returning all items in A that are also in B"""
    seen = set(map(hashable_repr, B))
    return tuple([n for n in A if hashable_repr(n) in seen])

def dict_intersect(A: dict, B: dict, match_vals= False) -> dict:
    """Returns a dict returning all items (keys:vals) in A that are also in B if match_vals is True
    else, only compares keys"""
    return dict({key:val for key, val in A.items() if (key, val) in B.items()}) if match_vals \
            else dict({key:A[key] for key in A if key in B})

def set_intersect(A: set, B: set) -> set:
    """Returns a set returning all members in A that are also in B"""
    return set(a for a in A if a in B)

def type_intersect(A: Iterables, B: Iterables, match_vals= False) -> Iterables:
    """General intersection for all types.  A and B must be the same type. Pass match_vals on to dicts"""
    if type(A) != type(B):
        raise TypeError("A and B must be the same type!")
    if isinstance(A, list) and isinstance(B, list):
        return list_intersect(A, B)
    elif isinstance(A, tuple) and isinstance(B, tuple):
        return tuple_intersect(A, B)
    elif isinstance(A, dict) and isinstance(B, dict):
        return dict_intersect(A, B, match_vals)
    elif isinstance(A, set) and isinstance(B, set):
        return set_intersect(A, B)
    else:
        raise TypeError(f"type_union(A: ${type(A)}, B: ${type(B)})\nWrong type signature")


#Syntax Converters
#----------------------------------------------------------------------------------------------------------------------

def tupler(a: tuple | Any) -> tuple:
    """
    Passes tuples and generator objects to tuples, and parenthesizes anything else.
    This effectively bypasses the need for a comma in defining a singleton tuple;  (a) -> (a,)
    """
    if isinstance(a, tuple | GeneratorType):
        return tuple(a)
    else:
        return (a,)


#Function Macros
#----------------------------------------------------------------------------------------------------------------------

def maybe_arg(func, pass_to_kwargs= False):
    """
    Usage: maybe_arg(func)(*args, **kwargs)        Reductive polymorphism.

    Modifies "func" to take an arbitrary number of args and kwargs but discards them (right to left) if "func"
    recieves too many.  Likewise, if a keyword argument not present in "func", it will be ignored.

    For example:
    def f(x, y, z=0)
    maybe_arg(f)(1,2,3) -> f(1,2)  #f ignores 3 because x and y are already satisfied.

    By default, any keyword argument will not accept positional arguments.  This can be changed with the
    "pass_to_kwargs" tag.  If set to True, positional arguments with a default value will also be filled.*

    *All positional kwargs will be overwritten by manually running with kwargs.
    """
    POSITIONAL_OR_KEYWORD = inspect.Parameter.POSITIONAL_OR_KEYWORD
    EMPTY = inspect.Parameter.empty
    VAR_POSITIONAL = inspect.Parameter.VAR_POSITIONAL

    params = inspect.signature(func).parameters.items()
    required_pars = tuple(par for _, par in params
                          if par.kind in (POSITIONAL_OR_KEYWORD, VAR_POSITIONAL) and par.default is EMPTY)

    def wrapper(*args, **kwargs):
        args = args if pass_to_kwargs or not required_pars or required_pars[-1].kind is VAR_POSITIONAL \
            else args[:len(required_pars)]
        try:
            return func(*args, **kwargs)
        except TypeError as e:
            s = str(e)
            if "positional argument" in s or "multiple values " in s:
                return wrapper(*args[:-1], **kwargs)
            elif "keyword argument '" in s:
                pos = s.index("keyword argument '") + 18
                k = s[pos:][:s[pos:].index("\'")]
                return wrapper(*args, **dict_compliment(kwargs,{k:kwargs[k]}))
            else:
                raise e

    return wrapper


def get_attributes(A: type) -> dict:
    """Return a dictionary of name-attribute pairs of a class"""
    return {name:getattr(A,name) for name in dir(A)}



#Functors
#---------------------------------------------------------------------------------------------------------------------

#TODO: UNTESTED! functor mapping of morphisms works for basic types.
class copy_type():
    """
    copy_type(basetype: type, name: str, attributes: dict) -> type

    Create a copy of a type into a new namespace.  If a namespace already exists, it will not create a new type but
    return the already cached type.
    """
    types = [int, float, str, list, tuple, dict, set, frozenset, bool, complex, bytes, bytearray, memoryview, type, object]
    type_cache = {t.__name__:t for t in types}

    def __new__(cls, basetype: type, name: str, attributes: dict | None = None) -> type:
        attributes = attributes if attributes is not None else {}
        if name not in cls.type_cache:
            #Default __repr__
            new_attributes = {"__repr__":lambda self: f"{name}({basetype(self)})"} | attributes

            #Create new type
            new_type = type(name, (basetype,), new_attributes)

            #Map reflexive morphisms to new type
            def wrap_end_type(f):
                def type_wrapper(*args, **kwargs):
                    result = f(*args, **kwargs)
                    return maybe_type(new_type, result) if type(result) is basetype else result

                return type_wrapper

            #Add more to this list if shit starts breaking
            exclude_attrs = dir(object) + list(attributes)  #Dont map user defined attributes!

            #setattrs for new_type
            for meth_name in dir(new_type):
                if meth_name not in exclude_attrs and callable(meth_attr := getattr(basetype, meth_name)):
                    setattr(new_type, meth_name, wrap_end_type(meth_attr))

            #Cache type
            cls.type_cache[name] = new_type

        return cls.type_cache[name]


def maybe_type(base_type: type, data: Any) -> Any:
    """Attempts to construct a type of 'base_type' with 'data' and returns 'data' if it fails."""
    try:
        return base_type(data)
    except TypeError:
        return data


def wrap_types(wrap: type, data: Any, base_types = ()) -> Any:
    """Attempts to apply a type constructor 'wrap' on 'data' if data is/in 'base_types'"""
    if type(base_types) is type:
        base_types = (base_types,)

    if type(data) in base_types:
        return maybe_type(wrap, data)

    else:
        return data




#Backwards Compatability #TODO: change search macros to own module?
from .searchmacros import replace_value_nested


from typing import Any, TypeVar
from .typemacros import tupler
from ._replace_value import _replace_value_recursive, _replace_value_mutable, BREAK_SEARCH
from copy import deepcopy




Iterables = TypeVar('Iterables', list, tuple, dict, set) 

def replace_value_nested(a: Iterables, old_vals: tuple | Any , new_val, callback = None, mode = 'return') -> Iterables:
    """
    Replaces a value(s) in a nested data structure.  The value(s) to replace can be of any type,
    including the type of the data structures being searched through.

    Use 'callback' to execute code on replace.
    Callback method should be in the form:  callback(old_val, new_val) -> new_val'
    The return of the method will be the new value.
    ('old_val' will be sampled from any matching value in 'old_vals')

    def callback(old, new):
        print(f"Value {old} replaced with {new}!")
        return new

    Return the BREAK_SEARCH object to end the search:

    def callback(old, new):
        global count
        count += 1
        return new if count <= 10 else BREAK_SEARCH

    Additionally, callback may accept a 'parents' keyword that references a list of all the sequential parents
    that the found value is nested in.
    'parents[0]' is the immediate parent, while 'parent[-1]' is the full structure.

    def callback(old, new, parents):
        global other_value
        if other_value in parents[0]:
            return new
        else:
            return old

    The 'mode' parameter switches between different algorithms:

        - 'return' (default)
            Uses a recursive find and replace algorithm and returns a new data structure without affecting the old one.
            Preserves child types but reinitializes them.  This algorithm is much slower than 'replace' and is subject
            to recursive depth limit.  Accessing 'parents' in the callback will never show updated values.

        - 'replace'
            Replaces the values in the structure and returns it.  Does not search through tuples and will keep
            all dependencies of child types.  Accessing 'parents' in the callback may yield values that have already
            been replaced.

        - 'copy'
            Same as 'replace' but creates a copy of the structure (fast, but high memory usage).  All internal
            dependencies will be conserved, but any references to data from outside the structure will not reference
            the cloned data.


    :param a: list, tuple, dict, or set  (all nestings allowed)
    :param old_vals: value(s) to replace
    :param new_val: value to inject
    :param callback: method::old_val -> new_val -> new_val'
    :param mode: switches between different algorithms
    :return: list, tuple, dict, or set
    """

    if mode == 'return':
        return _replace_value_recursive()(a, old_vals, new_val, callback)
    elif mode == 'replace':
        if isinstance(a, tuple):
            raise ValueError("Cannot replace a tuple")
        return _replace_value_mutable(a, old_vals, new_val, callback)
    elif mode == 'copy':
        if isinstance(a, tuple):
            raise ValueError("Cannot replace a tuple")
        new_a = deepcopy(a)
        return _replace_value_mutable(new_a, old_vals, new_val, callback)
    else:
        raise ValueError(f"{mode} is not a valid mode!\nValid modes: ['return', 'replace', 'copy']")



def find_nth(string: str, sub_str: str, n: int, idx = 0) -> int:
    """
    Returns the index of the nth occurance of sub_str in string. (0-based)
    Returns -1 if there is no nth occurance.
    """
    L = len(sub_str)
    index = string.find(sub_str)

    if index != -1:
        if n > 0:
            return find_nth(string[index + L:], sub_str, n - 1, idx + index + L)
        else:
            return idx + index
    else:
        return -1
